/*
 * Copyright (C) 2007 OpenedHand Ltd
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef HAVE_LAUNCHER_UTIL_H
#define HAVE_LAUNCHER_UTIL_H

#include <glib.h>
#include <gtk/gtk.h>

#include "taku-menu.h"

#define DESKTOP "Desktop Entry"

char **
exec_to_argv (const char *exec);

void
launcher_start (GtkWidget *widget,
                TakuMenuItem *item,
                gchar **argv,
                gboolean use_sn,
                gboolean single_instance);

GdkPixbuf*
get_icon (const gchar *icon_name, gint size);

#endif
